import asyncio
from itertools import islice
import os
from pathlib import Path
from typing import Iterable, Literal, Sequence


def longest_common_prefix(strings: list[str]) -> str:
    """
    Find the longest common prefix among a list of strings.

    Arguments:
        strings: List of strings

    Returns:
        The longest common prefix string
    """
    if not strings:
        return ""

    # Start with the first string as reference
    prefix: str = strings[0]

    # Compare with each subsequent string
    for current_string in strings[1:]:
        # Reduce prefix until it matches the start of current string
        while not current_string.startswith(prefix):
            prefix = prefix[:-1]
            if not prefix:
                return ""

    return prefix


class DirectoryReadTask:
    """A task to read a directory."""

    def __init__(self, path: Path) -> None:
        self.path = path
        self.done_event = asyncio.Event()
        self.directory_listing: list[Path] = []
        self._task: asyncio.Task | None = None

    def read(self) -> None:
        """Read the directory contents."""
        # TODO: Arbitary limit, to avoid stalling on very large directories
        # Should this be configurable?
        try:
            for path in islice(self.path.iterdir(), None, 10_000):
                self.directory_listing.append(path)
        except OSError:
            return

    def start(self) -> None:
        self._task = asyncio.create_task(
            self.run(), name=f"DirectoryReadTask({str(self.path)!r})"
        )

    async def run(self):
        try:
            await asyncio.to_thread(self.read)
        finally:
            self.done_event.set()

    async def wait(self) -> list[Path]:
        await self.done_event.wait()
        return self.directory_listing


class PathComplete:
    """Auto completes paths."""

    def __init__(self) -> None:
        self.read_tasks: dict[Path, DirectoryReadTask] = {}
        self.directory_listings: dict[Path, list[Path]] = {}

    @classmethod
    def decorate_listing(cls, paths: Iterable[Path]) -> list[str]:
        """Add trailing slash to directories.

        Args:
            paths: A sequence of paths.

        Returns:
            A list of directory names.
        """
        return [(path.name + "/" if path.is_dir() else path.name) for path in paths]

    @classmethod
    def decorate_path(cls, path: Path) -> str:
        """Add trailing slash to a path if it is a directory.

        Args:
            path: Path.

        Returns:
            The path name, potentially with a trailing slash.
        """
        return path.name + "/" if path.is_dir() else path.name

    async def __call__(
        self,
        current_working_directory: Path,
        path: str,
        *,
        exclude_type: Literal["file"] | Literal["dir"] | None = None,
    ) -> tuple[str | None, list[str] | None]:
        """Attempt to auto-complete a path.

        Args:
            current_working_directory: Current working directory (to resolve relative paths).
            path: String potentially containing path to auto-complete.
            exclude_type: May be `"file"` to exclude files, `"dir"` to exclude directories, or `None` not to exclude anything.

        Returns:
            A pair of the recommend auto complete (or `None`), and a list of options (or `None`).
        """

        current_working_directory = (
            current_working_directory.expanduser().resolve().absolute()
        )
        directory_path = (current_working_directory / Path(path).expanduser()).resolve()

        node: str = path
        if not directory_path.is_dir():
            node = directory_path.name
            directory_path = directory_path.parent

        if (listing := self.directory_listings.get(directory_path)) is None:
            read_task = DirectoryReadTask(directory_path)
            self.read_tasks[directory_path] = read_task
            read_task.start()
            listing = await read_task.wait()

        def check_completions(
            listing: list[Path],
        ) -> tuple[str | None, list[str] | None]:
            """Check completions from a listing (run in a thread)

            Args:
                listing: List of paths.

            Returns:
                Completions.
            """
            if exclude_type is not None:
                if exclude_type == "dir":
                    listing = [
                        listing_path
                        for listing_path in listing
                        if not listing_path.is_dir()
                    ]
                else:
                    listing = [
                        listing_path
                        for listing_path in listing
                        if listing_path.is_dir()
                    ]

            if not node:
                return None, self.decorate_listing(listing)

            node_name = Path(node).name

            if not node_name or path.endswith("/"):
                return None, self.decorate_listing(listing)
            matching_nodes = [
                listing_path
                for listing_path in listing
                if listing_path.name.startswith(node_name)
            ]

            if not (matching_nodes):
                # Nothing matches
                return None, None

            if not (
                prefix := longest_common_prefix(
                    [node_path.name for node_path in matching_nodes]
                )
            ):
                return None, None

            picked_path = directory_path / prefix

            complete_name = Path(path).name
            completed_prefix = prefix[len(complete_name) :]
            path_options = [
                (
                    matched_path.name[len(complete_name) :] + "/"
                    if matched_path.is_dir()
                    else matched_path.name[len(complete_name) :]
                )
                for matched_path in matching_nodes
                if matched_path.name.startswith(complete_name)
            ]

            if len(path_options) > 1:
                return None, path_options

            if picked_path.is_dir() and len(path_options) <= 1:
                completed_prefix += os.sep

            return completed_prefix or None, path_options

        return await asyncio.to_thread(check_completions, listing)


if __name__ == "__main__":

    async def run():
        path_complete = PathComplete()
        cwd = Path("~/sandbox")

        print(await path_complete(cwd, "~/p"))

    asyncio.run(run())



import httpx


# This doesn't work anonymously
# TODO: Implement login
def create_gist(
    files: dict[str, str],
    description: str = "",
    public: bool = True,
) -> str:
    """Create a GitHub Gist and return its URL.

    Args:
        files: A mapping of filename to file content.
        description: An optional description for the gist.
        public: Whether the gist should be public.

    Returns:
        The URL of the created gist.
    """
    payload = {
        "description": description,
        "public": public,
        "files": {name: {"content": content} for name, content in files.items()},
    }

    response = httpx.post(
        "https://api.github.com/gists",
        json=payload,
        headers={"Accept": "application/vnd.github+json"},
    )
    response.raise_for_status()
    return response.json()["html_url"]



import asyncio

from collections import Counter, defaultdict
from asyncio import Lock
from itertools import islice
from operator import itemgetter


class FuzzyIndex:
    """An index for path searching.

    This reduces the number of paths that need to be considered by the relativley expensive scoring function.

    """

    def __init__(self) -> None:
        self._paths: list[str] = []
        self._normalized_paths: list[str] = []
        self._index: dict[str, set[int]] = {}
        self._weights_cache: dict[str, dict[int, float]] = {}
        self._path_counter_cache: dict[str, Counter[str]] = {}
        self._lock = Lock()

    async def update_paths(self, paths: list[str]) -> None:
        """Update the paths and rebuild the index.

        Args:
            paths: New paths.
        """
        async with self._lock:
            self._paths = paths
            self._normalized_paths = await asyncio.to_thread(
                self._normalize_paths, paths
            )
            self._index = await asyncio.to_thread(
                self._build_trigram_index, self._normalized_paths
            )

    @classmethod
    def _normalize_paths(cls, paths: list[str]) -> list[str]:
        """Normalize the paths for searching.

        Args:
            paths: List of paths as strings.

        """
        normalized_paths = list(map(str.casefold, paths))
        return normalized_paths

    @classmethod
    def _extract_trigrams(cls, text: str) -> set[str]:
        """Return the set of all 3-character substrings (trigrams) in text.

        Padding with spaces ensures the start and end of the string are
        represented, so short strings and prefixes still produce trigrams.

        Args:
            text: candidate text (path).

        Returns:
            A set of 3 character strings.
        """
        padded = f"  {text} "
        return {padded[index : index + 3] for index in range(len(padded) - 2)}

    @classmethod
    def _build_trigram_index(cls, paths: list[str]) -> dict[str, set[int]]:
        """Build an inverted index mapping each trigram to the indices of paths containing it."""
        index: dict[str, set[int]] = defaultdict(set)
        for path_index, path in enumerate(paths):
            for trigram in cls._extract_trigrams(path):
                index[trigram].add(path_index)
        return index

    def _find_candidates(
        self,
        query: str,
        min_trigram_overlap: float = 0.3,
    ) -> list[tuple[str, str]]:
        """Return the indices of paths that share enough trigrams with the query.


        Args:
            query: The query to apply to the candidates.
            min_trigram_overlap: The overlap threshold (default 30%) controls the trade-off between
                recall (not missing real matches) and how many candidates get passed
                on to the slower detailed scoring stage.

        """
        # An upper limit to the number of candidates
        # If there are this many matches, the user will need to type an additional character or two
        MAX_CANDIDATES = 2000

        query = query.casefold()
        query_length = len(query)
        if query_length == 1:
            # One character
            # Find paths which have the query on the first component
            slash_query = f"/{query}"
            candidates = list(
                islice(
                    (
                        (path, normalized_path)
                        for path, normalized_path in zip(
                            self._paths, self._normalized_paths
                        )
                        if (
                            normalized_path.startswith(query)
                            or slash_query in normalized_path
                        )
                    ),
                    None,
                    MAX_CANDIDATES,
                )
            )
            return candidates

        if query_length <= 3:
            # Find paths which have all the characters
            query_set = set(query)
            candidates = list(
                islice(
                    (
                        (path, normalized_path)
                        for path, normalized_path in zip(
                            self._paths, self._normalized_paths
                        )
                        if query_set.issubset(normalized_path)
                    ),
                    None,
                    MAX_CANDIDATES,
                )
            )
            return candidates

        index = self._index
        query_trigrams = self._extract_trigrams(query)
        matching_buckets = [
            index[trigram] for trigram in query_trigrams if trigram in index
        ]

        trigram_match_counts = Counter(
            sorted(path_index for bucket in matching_buckets for path_index in bucket)
        )
        minimum_shared_trigrams = len(query_trigrams) * min_trigram_overlap
        unique_candidates = dict.fromkeys(
            islice(
                (
                    (self._paths[path_index], self._normalized_paths[path_index])
                    for path_index, count in trigram_match_counts.items()
                    if count >= minimum_shared_trigrams
                ),
                None,
                MAX_CANDIDATES,
            )
        )
        candidates = list(unique_candidates.keys())
        return candidates

    @classmethod
    def make_weights(cls, path: str) -> dict[int, float]:
        """Assign relative weights to positions with a path.

        Initial characters are weighted more highly, and the last component
        is weighted higher still.

        Args:
            path: A path string.

        Returns:
            A mapping of string indices to relative weights.
        """
        weights: dict[int, float] = dict.fromkeys(
            range(path.rfind("/") + 1, len(path)), 1.0
        ) | {0: 1.0}

        index = path.find("/", None, -1)
        while index != -1:
            weights[index + 1] = weights.setdefault(index + 1, 1.0) + 1.0
            index = path.find("/", index + 1, -1)
        return weights

    async def search(self, query: str) -> list[str]:

        normalized_query = query.casefold()
        query_counter = Counter(normalized_query)

        scores: list[tuple[float, str]] = []

        TOP_COUNT = 1000

        async with self._lock:
            for path, normalized_path in self._find_candidates(normalized_query):
                if (
                    path_counter := self._path_counter_cache.get(normalized_path)
                ) is None:
                    path_counter = self._path_counter_cache[normalized_path] = Counter(
                        normalized_path
                    )
                if query_counter - path_counter:
                    # Not all query characters are matched in the path
                    continue

                if (weights := self._weights_cache.get(normalized_path)) is None:
                    weights = self._weights_cache[normalized_path] = self.make_weights(
                        normalized_path
                    )

                score = sum(
                    [
                        weights.get(normalized_query.rfind(character), 0.0)
                        for character in normalized_path
                    ]
                )
                if normalized_path.endswith(normalized_query):
                    score *= 2
                if normalized_path.endswith(f"/{normalized_query}"):
                    score *= 2
                scores.append((score, path))

            scores.sort(reverse=True)
            top_scores = scores[:TOP_COUNT]

            return list(map(itemgetter(1), top_scores))


if __name__ == "__main__":

    # from textual._profile import timer

    from time import perf_counter
    import contextlib
    from typing import Generator

    @contextlib.contextmanager
    def timer(
        subject: str = "time", threshold: float = 0
    ) -> Generator[None, None, None]:
        """print the elapsed time. (only used in debugging).

        Args:
            subject: Text shown in log.
            threshold: Time in second after which the log is written.

        """
        start = perf_counter()
        yield
        elapsed = perf_counter() - start
        if elapsed >= threshold:
            elapsed_ms = elapsed * 1000
            print(f"{subject} elapsed {elapsed_ms:.4f}ms")

    async def run():
        with open("paths.txt") as f:
            PATHS = [line.rstrip() for line in f.readlines()]
        print(len(PATHS))
        # PATHS = [
        #     "~/foo/",
        #     "~/foo/bar/",
        #     "~/foo/bar/baz",
        #     "~/worldhello/foo",
        #     "~/Hello/world",
        # ]
        fuzzy = FuzzyIndex()
        with timer("update"):
            await fuzzy.update_paths(PATHS)
        with timer("search"):
            scores = await fuzzy.search("LICE")
        import rich

        rich.print(scores)

    import asyncio

    asyncio.run(run())



from importlib.resources import files
import asyncio

from toad.agent_schema import Agent


class AgentReadError(Exception):
    """Problem reading the agents."""


async def read_agents() -> dict[str, Agent]:
    """Read agent information from data/agents

    Raises:
        AgentReadError: If the files could not be read.

    Returns:
        A mapping of identity on to Agent dict.
    """
    import tomllib

    def read_agents() -> list[Agent]:
        """Read agent information.

        Stored in data/agents

        Returns:
            List of agent dicts.
        """
        agents: list[Agent] = []
        try:
            for file in files("toad.data").joinpath("agents").iterdir():
                agent: Agent = tomllib.load(file.open("rb"))
                if agent.get("active", True):
                    agents.append(agent)

        except Exception as error:
            raise AgentReadError(f"Failed to read agents; {error}")

        return agents

    agents = await asyncio.to_thread(read_agents)
    agent_map = {agent["identity"]: agent for agent in agents}

    return agent_map



from typing import Iterable

from textual.widgets import Markdown
from textual.widgets._markdown import MarkdownBlock


from toad.menus import MenuItem


class ConversationCodeFence(Markdown.BLOCKS["fence"]):

    def get_block_menu(self) -> Iterable[MenuItem]:
        yield from ()

    def get_block_content(self, destination: str) -> str | None:
        if destination == "clipboard":
            return self._content.plain
        return self.source


CUSTOM_BLOCKS = {"fence": ConversationCodeFence}


class ConversationMarkdown(Markdown):
    """Markdown widget with custom blocks."""

    def get_block_class(self, block_name: str) -> type[MarkdownBlock]:
        if (custom_block := CUSTOM_BLOCKS.get(block_name)) is not None:
            return custom_block
        return super().get_block_class(block_name)



# /// script
# dependencies = [
#   "agent-client-protocol"
# ]
# ///

import asyncio

from acp import (
    Agent,
    AgentSideConnection,
    InitializeRequest,
    InitializeResponse,
    NewSessionRequest,
    NewSessionResponse,
    PromptRequest,
    PromptResponse,
    session_notification,
    stdio_streams,
    text_block,
    update_agent_message,
)


class EchoAgent(Agent):
    def __init__(self, conn):
        self._conn = conn

    async def initialize(self, params: InitializeRequest) -> InitializeResponse:
        return InitializeResponse(protocolVersion=params.protocolVersion)

    async def newSession(self, params: NewSessionRequest) -> NewSessionResponse:
        return NewSessionResponse(sessionId="sess-1")

    async def prompt(self, params: PromptRequest) -> PromptResponse:
        for block in params.prompt:
            text = (
                block.get("text", "")
                if isinstance(block, dict)
                else getattr(block, "text", "")
            )
            await self._conn.sessionUpdate(
                session_notification(
                    params.sessionId,
                    update_agent_message(text_block(text)),
                )
            )
        return PromptResponse(stopReason="end_turn")


async def main() -> None:
    reader, writer = await stdio_streams()
    AgentSideConnection(lambda conn: EchoAgent(conn), writer, reader)
    await asyncio.Event().wait()


if __name__ == "__main__":
    asyncio.run(main())



import webbrowser
from urllib.parse import urlencode


def open_tweet_intent(
    text: str,
    *,
    url: str | None = None,
    hashtags: list[str] | None = None,
    via: str | None = None,
    in_reply_to: str | None = None,
) -> str:
    """Open a pre-filled Twitter/X tweet intent in the default browser.

    Args:
        text: The pre-filled tweet text.
        url: An optional URL to append to the tweet.
        hashtags: An optional list of hashtags to include (without '#').
        via: An optional username to attribute the tweet to.
        in_reply_to: An optional tweet ID to reply to.

    Returns:
        The full tweet intent URL that was opened.
    """
    params: dict[str, str] = {"text": text}

    if url is not None:
        params["url"] = url
    if hashtags:
        params["hashtags"] = ",".join(hashtags)
    if via is not None:
        params["via"] = via.lstrip("@")
    if in_reply_to is not None:
        params["in_reply_to"] = in_reply_to

    tweet_url = f"https://twitter.com/intent/tweet?{urlencode(params)}"
    webbrowser.open(tweet_url)
    return tweet_url



from __future__ import annotations

import asyncio
from itertools import filterfalse
from typing import Callable
from time import time
from os import PathLike
from pathlib import Path

from textual._partition import partition

from toad.path_filter import PathFilter


class ScanJob:
    """A single directory scanning job."""

    def __init__(
        self,
        name: str,
        queue: asyncio.Queue[Path],
        results: list[Path],
        path_filter: PathFilter | None = None,
        add_directories=False,
    ) -> None:
        self.queue = queue
        self.results = results
        self.name = name
        self.path_filter = path_filter
        self.add_directories = add_directories

    def start(self) -> None:
        self._task = asyncio.create_task(self.run())

    async def run(self) -> None:
        queue = self.queue
        results = self.results
        add_directories = self.add_directories
        while True:
            try:
                scan_path = await queue.get()
            except asyncio.QueueShutDown:
                break
            paths, dir_paths = await asyncio.to_thread(
                self._scan_directory, scan_path, self.path_filter
            )
            if add_directories:
                results.extend(dir_paths)
            results.extend(paths)
            try:
                for path in dir_paths:
                    await queue.put(path)
            except asyncio.QueueShutDown:
                break
            queue.task_done()

    def _scan_directory(
        self, root: Path, path_filter: PathFilter | None = None
    ) -> tuple[list[Path], list[Path]]:
        """Perform a directory scan (done in a thread).

        Args:
            root: Path to scan.
            path_filter: PathFilter object.

        Returns:
            A tuple of lists of paths (FILES, DIRECTORIES)
        """
        try:
            paths = list(root.iterdir())
        except IOError:
            paths = []
        if path_filter is not None:
            paths = list(filterfalse(path_filter.match, paths))

        try:
            paths, dir_paths = partition(Path.is_dir, paths)
        except IOError:
            paths = []
            dir_paths = []
        return paths, dir_paths


async def scan(
    root: Path,
    *,
    max_simultaneous: int = 5,
    path_filter: PathFilter | None = None,
    add_directories: bool = False,
    max_duration: float | None = 5.0,
) -> list[Path]:
    """Scan a directory for paths.

    Args:
        root: Root directory to scan.
        max_simultaneous: Maximum number of scan jobs.
        path_filter: Path filter object.
        add_directories: Also collect directories?
        max_duration: Maximum time in seconds to scan for, or `None` for no maximum.

    Returns:
        A list of Paths.
    """
    queue: asyncio.Queue[Path] = asyncio.Queue()
    results: list[Path] = []
    jobs = [
        ScanJob(
            f"scan-job #{index}",
            queue,
            results,
            path_filter=path_filter,
            add_directories=add_directories,
        )
        for index in range(max_simultaneous)
    ]
    try:
        await queue.put(root)
        for job in jobs:
            job.start()
        if max_duration is not None:
            try:
                async with asyncio.timeout(max_duration):
                    await queue.join()
            except asyncio.TimeoutError:
                pass
        else:
            await queue.join()
    except asyncio.CancelledError:
        await queue.join()
    queue.shutdown(immediate=True)
    return results


class Scan:
    """A scan of a single directory."""

    def __init__(self, root: Path, on_complete: Callable[[Scan]]) -> None:
        self.root = root
        self._on_complete = on_complete
        self._complete_event = asyncio.Event()
        self._scan_result: list[Path] = []
        self._scan_task: asyncio.Task | None = None
        self._scan_time = time()

    @property
    def is_complete(self) -> bool:
        """Has the scan finished?"""
        return self._complete_event.is_set()

    def start(self) -> None:
        self._scan_time = time()
        self._scan_task = asyncio.create_task(self._run(), name=f"scan {self.root!s}")

    async def _run(self) -> None:
        await asyncio.to_thread(self._scan)
        self._complete_event.set()
        self._on_complete(self)

    def _scan(self) -> None:
        self._scan_result = list(self.root.iterdir())

    async def wait(self) -> list[Path]:
        """Get the result of the scan, potentially waiting for it to finish first.

        Returns:
            A list of paths in the root.
        """
        await self._complete_event.wait()
        assert self._scan_result is not None
        self._scan_task = None
        return self._scan_result


class DirectoryScanner:
    """Object to recursively scan a directory."""

    def __init__(self, root: PathLike) -> None:
        self.root = Path(root)
        self.directories: dict[Path, Scan] = {}

    async def scan(
        self, relative_directory_path: str, on_complete: Callable[[Scan]]
    ) -> Scan:
        """Get a scan.

        Scans are created on demand, or returned previously scanned.

        Args:
            relative_directory_path: A path relative to the root.
            on_complete: Callback when scan is complete, will be invoked with Scan instance.

        Returns:
            A scan instance.
        """
        scan_path = self.root / relative_directory_path
        if scan := self.directories.get(scan_path):
            if scan.is_complete:
                on_complete(scan)
        else:
            self.directories[scan_path] = scan = Scan(
                scan_path, on_complete=on_complete
            )
            scan.start()
        return scan


if __name__ == "__main__":
    import asyncio

    import contextlib
    from time import perf_counter
    from typing import Generator

    @contextlib.contextmanager
    def timer(subject: str = "time") -> Generator[None, None, None]:
        """print the elapsed time. (only used in debugging)"""
        start = perf_counter()
        yield
        elapsed = perf_counter() - start
        elapsed_ms = elapsed
        print(f"{subject} elapsed {elapsed_ms:.4f}s")

    from toad.path_filter import PathFilter

    scan_path = Path("~/projects/textual").expanduser()

    path_filter = PathFilter.from_git_root(scan_path)

    async def run():
        with timer("scan"):
            return await scan(scan_path, path_filter=path_filter)

    paths = asyncio.run(run())
    print(len(paths))



from textual.app import ComposeResult
from textual import getters

from toad.app import ToadApp
from toad.widgets.grid_select import GridSelect
from toad.widgets.session_summary import SessionSummary
from toad.session_tracker import SessionTracker, SessionDetails


class SessionGridSelect(GridSelect):
    FOCUS_ON_CLICK = True
    app: getters.app[ToadApp] = getters.app(ToadApp)

    def __init__(
        self,
        session_tracker: SessionTracker,
        id: str | None = None,
        classes: str | None = None,
    ) -> None:
        self.session_tracker = session_tracker
        super().__init__(
            id=id,
            classes=classes,
            min_column_width=36,
        )

    def allow_focus(self) -> bool:
        return True

    def on_mount(self) -> None:
        self.app.session_update_signal.subscribe(
            self, self.handle_session_update_signal
        )

    def update_current(self, current_mode: str) -> None:
        for index, session_summary in enumerate(self.query_children(SessionSummary)):
            session_summary.current = False
            if session_summary.session_details is not None:
                current = session_summary.session_details.mode_name == current_mode
                session_summary.current = current
                if current:
                    self.highlighted = index
                    break

    async def handle_session_update_signal(
        self, update: tuple[str, SessionDetails | None]
    ) -> None:
        mode_name, details = update
        session_summary = self.query_one_optional(f"#{mode_name}", SessionSummary)
        if details is None:
            if session_summary is not None:
                await session_summary.remove()
            return

        if session_summary is None:
            await self.mount(SessionSummary(details, id=details.mode_name))
        else:
            session_summary.session_details = details

    def compose(self) -> ComposeResult:
        for session in self.session_tracker.ordered_sessions:
            yield SessionSummary(session, id=session.mode_name)



from dataclasses import dataclass
from operator import attrgetter
from typing import Iterable, Literal, Sequence

from textual.signal import Signal

type SessionState = Literal["notready", "busy", "asking", "idle"]


@dataclass
class SessionDetails:
    """Tracks a concurrent session."""

    index: int
    """Index of session, used in sorting."""
    mode_name: str
    """The screen mode name."""
    title: str = ""
    """The title of the conversation."""
    subtitle: str = ""
    """The subtitle of the conversation."""
    path: str = ""
    """The project directory path."""
    state: SessionState = "notready"
    """The current state of the session."""
    summary: str = ""
    """Suplimentary information about the session."""

    updates: int = 0
    """Track updates to the session details."""


class SessionTracker:
    """Tracks concurrent agent settings"""

    def __init__(self, signal: Signal[tuple[str, SessionDetails | None]]) -> None:
        self.sessions: dict[str, SessionDetails] = {}
        self._session_index = 0
        self.signal = signal

    @property
    def session_count(self) -> int:
        return len(self.sessions)

    def new_session(self) -> SessionDetails:
        self._session_index += 1
        mode_name = f"session-{self._session_index}"
        session_meta = SessionDetails(
            index=self._session_index, mode_name=mode_name, title="New Session"
        )
        self.sessions[mode_name] = session_meta
        return session_meta

    def close_session(self, mode_name: str) -> None:
        if mode_name in self.sessions:
            del self.sessions[mode_name]
            self.signal.publish((mode_name, None))

    def get_session(self, mode_name: str) -> SessionDetails | None:
        return self.sessions.get(mode_name, None)

    def update_session(
        self,
        mode_name: str,
        title: str | None = None,
        subtitle: str | None = None,
        path: str | None = None,
        state: SessionState | None = None,
    ) -> SessionDetails:
        session_details = self.sessions[mode_name]
        if title is not None:
            session_details.title = title
        if subtitle is not None:
            session_details.subtitle = subtitle
        if path is not None:
            session_details.path = path
        if state is not None:
            session_details.state = state
        self.signal.publish((mode_name, session_details))
        return session_details

    @property
    def ordered_sessions(self) -> Sequence[SessionDetails]:
        return sorted(self.sessions.values(), key=attrgetter("index"))

    def __iter__(self) -> Iterable[SessionDetails]:
        return iter(self.ordered_sessions)

    def session_cursor_move(
        self, mode_name: str, direction: Literal[-1, +1]
    ) -> str | None:
        mode_names = [session.mode_name for session in self.ordered_sessions]
        try:
            mode_index = mode_names.index(mode_name)
        except ValueError:
            return None
        mode_index = (mode_index + direction) % len(mode_names)
        new_mode_name = mode_names[mode_index]
        return new_mode_name
